import logging
import os
from typing import Optional, Union, Tuple

from humanfriendly import format_size

from dslib.utils import timeit

try:
    import pandas as pd
    from google.cloud import bigquery
    from google.api_core.exceptions import BadRequest, GoogleAPIError
    from google.cloud.bigquery_storage import BigQueryReadClient
except ImportError as e:
    raise ImportError('Please install "google" extra packages.') from e

# Set up logger
LOGGER = logging.getLogger(__name__)
logging.getLogger('google.auth._default').setLevel(logging.ERROR)  # Silence google auth warning/info/debug logs
logging.getLogger('pandas_gbq.gbq').setLevel(logging.WARNING)  # Silence pandas GBQ info/debug logs

# Set up special types
JobType = bigquery.job._AsyncJob


class BigQueryError(Exception):
    pass


class BigQueryWrapper(object):

    def __init__(self, client: bigquery.Client, bqstorage_client: BigQueryReadClient = None) -> None:
        if not client.__class__.__name__ == 'Client':
            raise ValueError('"client" argument must be a Google "Client" instance')
        self.client = client
        self._bqstorage_client = bqstorage_client

    @classmethod
    def from_params(cls, project: str, auth_file: Optional[str] = None, setup_bqstorage: bool = True):
        if auth_file:
            if not os.path.exists(auth_file):
                raise FileNotFoundError(f'Authentication file does not exist: {auth_file}')
            client = bigquery.Client.from_service_account_json(auth_file, **{'project': project})
            bqstorage_client = BigQueryReadClient.from_service_account_json(auth_file) if setup_bqstorage else None
        else:
            client = bigquery.Client(project=project)
            bqstorage_client = BigQueryReadClient() if setup_bqstorage else None
        return BigQueryWrapper(client, bqstorage_client)

    @property
    def bqstorage_client(self):
        if self._bqstorage_client is None:
            raise BigQueryError('BigQuery Storage was not set up for this wrapper')
        return self._bqstorage_client

    def assess_query(self, query: str) -> bool:
        """
        Assesses whether the query is valid or not. Logs the volume (in bytes) of data it would process.
        :param query: query to be executed
        :return: bool
        """
        LOGGER.info('Dry-running query...')
        job_config = bigquery.QueryJobConfig()
        job_config.dry_run = True
        job_config.use_query_cache = False
        try:
            query_job = self.client.query(query, job_config=job_config)
        except BadRequest as e:
            LOGGER.error(f'Query is not valid: \n{str(e)}')
            return False
        else:
            LOGGER.info(f'Query is valid. It would process {format_size(query_job.total_bytes_processed)}')
            return True

    def run_query(self, query: str, output_type: Union[str, None] = 'df', timeout: Optional[int] = None,
                  tags: Optional[dict] = None, use_query_cache: bool = True, use_progress_bar: bool = True,
                  use_bq_storage_api: bool = True) -> Union[pd.DataFrame, None]:
        """
        Runs a query and returns the output in the specified format
        :param query: query to be executed
        :param output_type: output type of the run_query function. One of ['df', None]. Defaults to 'df'
        :param timeout: maximum allowed for query to run (in seconds). If None, no timeout is set
        :param tags: dictionary with "type", "category" and "sub_category" keys to handle query tagging,
            e.g. {'type': 'reporting', 'category': 'tableau-folder', 'sub_category': 'workbook-name'}
        :param use_query_cache: whether to use query's cached results when queried tables have not changed
        :param use_progress_bar: whether to use a tqdm progress bar during fetching. Defaults to True
        :param use_bq_storage_api: whether to use BigQuery Storage API to speed up fetching output. Defaults to True
        :return: results in the chosen output_type
        """
        # If tags are specified, add header to query
        query = self._add_tags(query, tags)
        # Start query run
        LOGGER.info('Running query...')
        query_job = self._start_query(query, use_query_cache=use_query_cache)
        # Wait during query execution
        query_job = self._wait_until_finished(query_job, timeout)
        # Get query output
        results = self._fetch_output(query_job, output_type, use_progress_bar, use_bq_storage_api)
        LOGGER.info(f'Query processed: {format_size(query_job.total_bytes_processed)}')
        return results

    @staticmethod
    def _add_tags(query: str, tags: Optional[dict] = None) -> str:
        if tags:
            head = f"/* type='{tags['type']}', category='{tags['category']}', sub_category='{tags['sub_category']}' */"
            return head + '\n' + query
        return query

    def _start_query(self, query: str, use_query_cache: bool = True, to_table: Optional[str] = None,
                     partition_by: Optional[str] = None, if_exists: str = 'fail') -> JobType:
        job_config = bigquery.QueryJobConfig()
        if not use_query_cache:
            job_config.use_query_cache = False
        if to_table:
            _, dataset, table = self._parse_table_reference(to_table)
            job_config.destination = self.client.dataset(dataset).table(table)
            job_config.write_disposition = self._get_write_disposition(if_exists)
            if partition_by:
                job_config = self._add_partitioning(job_config, partition_by)
        else:
            if partition_by:
                raise BigQueryError('Argument "partition_by" can only be used when "to_table" is provided')
        return self.client.query(query, job_config=job_config)

    @staticmethod
    def _get_write_disposition(if_exists: str) -> str:
        if if_exists == 'fail':
            return bigquery.WriteDisposition.WRITE_EMPTY
        elif if_exists == 'append':
            return bigquery.WriteDisposition.WRITE_APPEND
        elif if_exists == 'replace':
            return bigquery.WriteDisposition.WRITE_TRUNCATE
        else:
            raise NotImplementedError('"if_exists" argument must be in ("fail", "append", "replace")')

    def _parse_table_reference(self, table_reference: str) -> Tuple:
        table_reference_split = table_reference.split('.')
        if len(table_reference_split) == 2:
            return self.client.project, table_reference_split[0], table_reference_split[1]
        elif len(table_reference_split) == 3:
            return tuple(table_reference_split)
        else:
            raise ValueError('Table ref must be in format: "dataset_id.table_id" or "project_id.dataset_id.table_name"')

    @staticmethod
    def _parse_table_schema(table_schema: dict) -> list:
        return [bigquery.SchemaField(column_name, bq_type) for column_name, bq_type in table_schema.items()]

    @staticmethod
    def _add_partitioning(job_config, partition_by):
        job_config.time_partitioning = bigquery.TimePartitioning(
            type_=bigquery.TimePartitioningType.DAY,
            field=partition_by
        )
        return job_config

    @staticmethod
    def _wait_until_finished(query_job: JobType, timeout: Optional[int] = None) -> JobType:
        with timeit(text='Query run took:', logger=LOGGER):
            try:
                query_job.result(timeout=timeout)  # Raises when errored or timed out
            except GoogleAPIError as exception:
                raise BigQueryError(exception.errors[0]['message']) from exception
        return query_job

    def _fetch_output(self, query_job: JobType, output_type: str, use_progress_bar: bool = True,
                      use_bq_storage_api: bool = True) -> Union[pd.DataFrame, None]:
        if output_type is None:
            return None
        elif output_type == 'df':
            result = query_job.result()
            with timeit(text=f'Fetching output "df" took:', logger=LOGGER):
                params = {}
                if use_progress_bar:
                    params['progress_bar_type'] = 'tqdm'
                if use_bq_storage_api:
                    params['bqstorage_client'] = self.bqstorage_client
                df = result.to_arrow(**params).to_pandas()
                LOGGER.info(f'Output "df" will take {format_size(df.memory_usage().sum())} in memory')
                return df
        else:
            raise NotImplementedError('Argument output_type must be in: ["df", None]')

    @timeit(text='Uploading df to table took:', logger=LOGGER)
    def to_table_from_df(self, df: pd.DataFrame, table_reference: str, if_exists: str = 'fail',
                         chunksize: Optional[int] = None, use_progress_bar: bool = True) -> None:
        """
        Uploads a DataFrame to a BigQuery table and creates it if it does not exist yet.
        :param df: pandas DataFrame to upload to BigQuery
        :param table_reference: reference of the table to create/append/replace (e.g. "warehouse.trip_fill" or
            "bbc-data-platform.warehouse.trip_fill")
        :param if_exists: action to take if the table already exists. Should be in ('fail', 'append', 'replace')
        :param chunksize: number of rows to be inserted in each chunk from the dataframe. Defaults to None, which loads
            the whole dataframe at once
        :param use_progress_bar: whether to use a tqdm progress bar during fetching. Defaults to True
        :return: None
        """
        LOGGER.info(f'Uploading df to {table_reference}...')
        project, dataset, table = self._parse_table_reference(table_reference)
        df.to_gbq(f'{dataset}.{table}', project_id=project, if_exists=if_exists, chunksize=chunksize,
                  progress_bar=use_progress_bar, credentials=self.client._credentials)

    @timeit(text='Running query and storing results into table took:', logger=LOGGER)
    def to_table_from_query(self, query: str, table_reference: Optional[str] = None, if_exists: str = 'fail',
                            tags: Optional[dict] = None, use_query_cache: bool = True,
                            partition_by: Optional[str] = None) -> None:
        """
        Run a query and load its result into a table.
        :param query: query to be executed
        :param table_reference: reference of the table to create/append/replace (e.g. "dataset_id.table_name" or
            "project_id.dataset_id.table_name")
        :param if_exists: behaviour to adopt when uploading results to BigQuery. Ignored when `to_table` is not
            provided. If provided, must be in ('fail', 'append', 'replace'). Defaults to 'fail'
        :param tags: dictionary with "type", "category" and "sub_category" keys to handle query tagging,
            e.g. {'type': 'reporting', 'category': 'tableau-folder', 'sub_category': 'workbook-name'}
        :param use_query_cache: whether to use query's cached results when queried tables have not changed
        :param partition_by: name of the date column on which to partition the created table by
        :return: None
        """
        # If tags are specified, add header to query
        query = self._add_tags(query, tags)
        # Run the query and load its result to a table
        LOGGER.info('Running query...')
        query_job = self._start_query(query, use_query_cache, to_table=table_reference, if_exists=if_exists,
                                      partition_by=partition_by)
        # Wait until query is finished
        self._wait_until_finished(query_job)  # Raises when errored
        LOGGER.info(f'Query processed: {format_size(query_job.total_bytes_processed)}')

    @timeit(text='Loading GCS data to BigQuery took:', logger=LOGGER)
    def to_table_from_gcs(self, source_gcs_uri: str, table_reference: str, source_type: str = 'csv',
                          if_exists: str = 'fail', table_schema: Optional[dict] = None,
                          partition_by: Optional[str] = None) -> None:
        """
        Load one or more GCS file(s) in BigQuery.
        :param source_gcs_uri: URI to  the file(s) to be loaded in BigQuery, e.g. 'gs://bucket_name/file_part_*.csv'.
            Compression in gzip is inferred from files and supported.
        :param table_reference: reference of the table to create/append/replace (e.g. "dataset_id.table_name" or
            "project_id.dataset_id.table_name")
        :param source_type: type of the files to be loaded in BigQuery. Must be in ("csv", "json"). Defaults to 'csv'
        :param if_exists: behaviour to adopt when uploading results to BigQuery. Ignored when `to_table` is not
            provided. If provided, must be in ('fail', 'append', 'replace'). Defaults to 'fail'
        :param table_schema: optional dictionary mapping the column names to be used to their associated BigQuery type,
            e.g. {'col_1': 'STRING', 'col_2': 'INT64', 'col_3': 'TIMESTAMP'}
        :param partition_by: name of the date column on which to partition the created table by
        :return: None
        """
        # Parse table reference
        project, dataset, table = self._parse_table_reference(table_reference)
        destination_table = bigquery.DatasetReference(project, dataset).table(table)
        # Create job config
        job_config = bigquery.LoadJobConfig()
        job_config.write_disposition = self._get_write_disposition(if_exists)
        # Handle table schema
        if table_schema:
            job_config.schema = self._parse_table_schema(table_schema)
        else:
            job_config.autodetect = True
        # Handle partitioning
        if partition_by is not None:
            job_config = self._add_partitioning(job_config, partition_by)
        # Handle source format
        if source_type == 'csv':
            job_config.source_format = bigquery.SourceFormat.CSV
            if table_schema:
                job_config.skip_leading_rows = 1
        elif source_type == 'json':
            job_config.source_format = bigquery.SourceFormat.NEWLINE_DELIMITED_JSON
        else:
            raise NotImplementedError('Argument source_type must be in: ["csv", "json"]')
        # Start table load job
        load_job = self.client.load_table_from_uri(source_gcs_uri, destination_table, job_config=job_config)
        LOGGER.info(f'Loading CSV from {source_gcs_uri} to {table_reference}')
        # Wait for table load to complete
        self._wait_until_finished(load_job)  # Raises when errored



from typing import Optional

import pandas as pd

from dslib.utils import deep_suppress_stdout_stderr

try:
    from fbprophet import Prophet
except ImportError as e:
    raise ImportError('Please install "prophet" extra packages.') from e


class FastProphet(Prophet):
    """
    Model using FB-Prophet but without the (slow) Monte-Carlo analysis used to get the confidence
    intervals.
    """

    def fit(self, df: pd.DataFrame, **kwargs) -> None:
        with deep_suppress_stdout_stderr():
            super(FastProphet, self).fit(df, **kwargs)

    def predict(self, X_test: Optional[pd.DataFrame] = None) -> pd.DataFrame:

        df = X_test if X_test is not None else None
        with deep_suppress_stdout_stderr():
            if df is None:
                df = self.history.copy()
            else:
                if df.shape[0] == 0:
                    raise ValueError('Dataframe has no rows.')
                df = self.setup_dataframe(df.copy())

            df['trend'] = self.predict_trend(df)
            seasonal_components = self.predict_seasonal_components(df)

            # Drop columns except y, ds, cap, floor, and trend
            cols = ['ds', 'trend']
            if 'cap' in df:
                cols.append('cap')
            if self.logistic_floor:
                cols.append('floor')
            # Add in forecast components
            df2 = pd.concat((df[cols], seasonal_components), axis=1)
            df2['yhat'] = (df2['trend'] * (1 + df2['multiplicative_terms']) + df2['additive_terms'])

            return df2



import pandas as pd

from dslib.utils import log


def configure_pandas(max_rows=None, max_columns=None, max_width=None, max_colwidth=None, logger=None):
    pd.set_option('display.max_rows', max_rows)
    pd.set_option('display.max_columns', max_columns)
    pd.set_option('display.max_colwidth', max_colwidth)
    pd.set_option('display.width', max_width)
    pd.set_option('display.expand_frame_repr', True)
    log('Pandas options were successfully set', logger=logger)



from ._caching import timed_cache
from ._io import ensure_directory_exists, ignore_exceptions, safe_write, read_yaml, get_tmp_dirpath, get_tmp_filepath, \
    zip_files, unzip_file
from ._logging import configure_logging, log, timeit, suppress_stdout, deep_suppress_stdout_stderr
from ._meta import resolve, compose
from ._templating import render_file, render_string
from ._temporality import midpoint_date, date_range, change_timezone, force_timezone_onto_timestamp



import logging
import numpy as np
import os

from dslib.utils import log
from typing import Optional

try:
    import dill
    from sklearn.pipeline import Pipeline
    from sklearn.base import TransformerMixin, BaseEstimator
    from sklearn.metrics import roc_auc_score, average_precision_score, recall_score, precision_score, precision_recall_curve
    from matplotlib import pyplot as plt
except ImportError:
    raise ImportError(f'Please install "ml" extra package.')


def predict_with_threshold(clf, df, threshold):
    """
    Perform class prediction with given trained binary classifier and threshold
    :param clf: binary classifier to be used
    :param df: input dataframe to predict on
    :param threshold: threshold to be used for classification
    :return: the classes predicted
    """
    # Assert threshold value is valid
    assert 0 < threshold < 1, 'Your threshold must be in interval ]0,1['
    # Use the model to predict probabilities
    y_prob = clf.predict_proba(df)[:, 1]
    # Choose a class based on the probabilities and the threshold
    return np.abs(np.round(y_prob - (threshold - 0.5)))  # Cut-off moved from 0.5 to threshold


def eval_score(y_true: list, y_pred: list, plot_pr_curve: bool = True,
               threshold: bool = None, logger: Optional[logging.Logger] = None):
    """
    Returns ROC and average precision score as well as PR curve given a vector of scores and labels.
    Give precision and recall if a threshold is provided
    :param y_true: label
    :param y_pred: prediction score
    :param plot_pr_curve: if set to True, will display precision recall (PR) curve (optional - default: True)
    :param threshold: threshold (optional)
    :param logger: logger to use to output results. If None, will use print (optional - default: None)
    """
    auc = roc_auc_score(y_true, y_pred)
    aps = average_precision_score(y_true, y_pred)
    log(f'AUC = {auc:.5}', logger)
    log(f'Average precision score = {aps:.5}', logger)
    log(f'Data size: {len(y_true)}', logger)

    if plot_pr_curve:
        p_ret, r_ret, t_ret = precision_recall_curve(y_true, y_pred)
        plt.plot(r_ret, p_ret)

        plt.legend()
        plt.title('Precision-recall curve')
        plt.xlabel('Recall')
        plt.ylabel('Precision')

    if threshold:
        log(f'Precision: {precision_score(y_true, y_pred > threshold):.5}', logger)
        log(f'Recall: {recall_score(y_true, y_pred > threshold):.5}', logger)


def eval_pipeline(pipeline: Pipeline, X, y_true: list, plot_pr_curve: bool = True, threshold: float = None):
    """
    Evaluate a pipeline using the eval_score method
    :param pipeline to evaluate
    :param X: input test dataset
    :param y_true: label to predict
    :param plot_pr_curve: if set to True, will display precision recall (PR) curve (optional - default: True)
    :param threshold: to use for classification
    """
    y_pred = pipeline.predict_proba(X)[:, 1]
    eval_score(y_true, y_pred, plot_pr_curve, threshold)


class FeatureSelection(BaseEstimator, TransformerMixin):
    """
    Class that performs Feature Selection as  part of a Sklearn's pipeline. Can work
    both as Feature Selection and Feature "dropping"
    """
    def __init__(self, feature_list: list, exclude=False):
        """
        exclude: optional (default = False). If set to True, the list of features given will be dropped instead
        of kept.
        """
        self.feature_list = feature_list
        self.exclude = exclude

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        if self.exclude:
            return X.drop(self.feature_list, axis=1)
        return X[self.feature_list]


def load_dill_model(model_path: str):
    """
    Load model which was saved in a dill format
    :param model_path: path of where to find the model
    """
    with open(model_path, 'rb') as model_file:
        u = dill.Unpickler(model_file)
        u.encoding = 'utf-8'
        return u.load()


def save_dill_model(model_path: str, pipeline, if_exists_replace: bool = False,
                    logger: Optional[logging.Logger] = None):
    """
    Save model in dill format
    :param model_path: path where to store the model
    :param pipeline: object to store
    :param if_exists_replace: boolean. If set to True, will replace file if already exists. If False, will raise
    Error (optional - default False)
    :param logger: logger to use to output results. If None, will use print (optional - default: None)
    """

    if not if_exists_replace:
        if os.path.isfile(model_path):
            raise FileExistsError(
                f"File {model_path} already exists. To overwrite it, set param if_exists_replace to True")

    with open(model_path, 'wb') as model_file:
        u = dill.Pickler(model_file, recurse=True)
        u.encoding = 'utf-8'
        u.dump(pipeline)
    log(f'Model saved at {model_path}', logger)



from datetime import datetime, timedelta
import functools


def timed_cache(maxsize: int = 128, typed: bool = False, **timed_cache_kwargs):
    """
    Decorator based on functools.lru_cache to add a cache expiration time
    Supports the arguments of functools.lru_cache and any combination of datetime.timedelta's arguments
    Credit to https://gist.github.com/Morreski/c1d08a3afa4040815eafd3891e16b945

    :param maxsize: Maximum number of distinct entries in the cache dictionary
    :param typed: Whether to consider two values identical but with different types as equal or not
    :param timed_cache_kwargs:
        microseconds: float
        milliseconds: float
        seconds: float
        minutes: float
        hours: float
        days: float
        weeks: float

    Example of usage:
    @timed_cache(minutes=10)
    def my_fcn():
        pass
    """
    def _wrapper(f):
        update_delta = timedelta(**timed_cache_kwargs)
        next_update = datetime.utcnow() - update_delta
        f = functools.lru_cache(maxsize=maxsize, typed=typed)(f)

        @functools.wraps(f)
        def _wrapped(*args, **kwargs):
            nonlocal next_update
            now = datetime.utcnow()
            if now >= next_update:
                f.cache_clear()
                next_update = now + update_delta
            return f(*args, **kwargs)

        return _wrapped

    return _wrapper



from functools import reduce
from typing import Any, Callable


def resolve(name: str) -> Any:
    """
    Loads a variable, function, class or module from a given dotted path (e.g. 'dslib.utils.resolve').
    Copied directly from: https://github.com/python/cpython/blob/master/Lib/logging/config.py
    :param name: (str) name (or dotted path) of the element to be imported
    :return: loaded class, function or module
    """
    name = name.split('.')
    used = name.pop(0)
    found = __import__(used)
    for n in name:
        used = used + '.' + n
        try:
            found = getattr(found, n)
        except AttributeError:
            __import__(used)
            found = getattr(found, n)
    return found


def _compose_2(f: Callable, g: Callable) -> Callable:
    return lambda *a, **kw: f(g(*a, **kw))


def compose(*functions: Callable) -> Callable:
    """
    Composes functions provided as input and returns the composite function (e.g. compose(f, g, h) will return fogoh)
    :param functions: (func) input functions to be composed
    :return: (func) composite function
    """
    return reduce(_compose_2, functions)



import logging
import warnings
from abc import abstractmethod
from typing import Optional, Union, Sequence, List, Dict, Tuple

from dslib.utils import timeit

try:
    import pandas as pd
    import psycopg2
    import pymysql
    from pandas.io.sql import pandasSQL_builder, _wrap_result
    from sqlalchemy import create_engine
    from sqlalchemy.exc import SQLAlchemyError
except ImportError as e:
    raise ImportError('Please install "database" extra packages.') from e

# Set up logger
LOGGER = logging.getLogger(__name__)
warnings.filterwarnings('ignore', message=r'\(1003\,')  # Hide pymysql 1003 persistent warning


class _DatabaseWrapper(object):

    def __init__(self, conn, expected_db_type):
        if not conn.__class__.__name__ == 'Engine':
            raise ValueError('"conn" argument must be a SQLAlchemy "Engine" instance')
        if not expected_db_type == conn.name:
            ValueError(f'Provided connector should be "{expected_db_type}" (current: "{conn.name}")')
        self.conn = conn
        self.db_type = conn.name
        self._pandas_sql = pandasSQL_builder(self.conn)

    @classmethod
    @abstractmethod
    def from_params(cls, *args, **kwargs):
        raise NotImplementedError

    def _execute(self, query, *args, **kwargs):
        return self._pandas_sql.execute(query, *args, **kwargs)

    def assess_query(self, query: str) -> bool:
        """
        Assesses whether the query is valid or not.
        :param query: query to be executed
        :return: bool
        """
        try:
            self._execute(f'EXPLAIN {query}', output_type=None)
        except SQLAlchemyError as e:
            LOGGER.error(f'Query is not valid: \n{str(e)}')
            return False
        else:
            LOGGER.info(f'Query is valid')
            return True

    def run_query(self, query: str, output_type: Optional[str] = 'df', index_col: Union[str, List[str], None] = None,
                  coerce_float: bool = True, parse_dates: Union[List[str], Dict[str, str], None] = None
                  ) -> Union[pd.DataFrame, None]:
        """
        Runs a query and returns the output in the specified format
        :param query: query to be executed
        :param output_type: output type of the run_query function. One of ['df', None]. Defaults to 'df'
        :param index_col: column(s) to set as index(MultiIndex)
        :param coerce_float: attempts to convert values of non-string, non-numeric objects (like decimal.Decimal) to
            floating point. Can result in loss of Precision
        :param parse_dates: either a list of column names to parse as dates, a dict of ``{column_name: format string}``,
            or a dict of ``{column_name: arg dict}``, where the arg dict corresponds to the keyword arguments of
            `pandas.to_datetime` function
        :return: results in the chosen output_type
        """
        # Execute query
        with timeit(text='Executing query took:', logger=LOGGER):
            result = self._execute(query)
        # Return if output_type is None
        if output_type is None:
            return
        # Else return the result as a Pandas DF
        elif output_type == 'df':
            with timeit(text='Fetching df took:', logger=LOGGER):
                columns = result.keys()
                data = result.fetchall()
                f = _wrap_result(data, columns, index_col=index_col, coerce_float=coerce_float, parse_dates=parse_dates)
                return f
        else:
            raise NotImplementedError('Argument output_type must be in: ["df", None]')

    @staticmethod
    def _parse_table_reference(table_reference: str) -> Tuple:
        table_name_split = table_reference.split('.')
        if len(table_name_split) == 1:
            return None, table_reference
        elif len(table_name_split) == 2:
            return tuple(table_name_split)
        else:
            raise ValueError('Table reference must be in format: "table_name" or "schema_name.table_name"')

    def exists(self, table_name: str, schema: str = None) -> bool:
        """"
        Check whether a table already exists or not.
        :param table_name: name of the table to check the existence of
        :param schema: name of the schema of the table to check the existence of
        :return: bool
        """
        return self.conn.dialect.has_table(self.conn, table_name, schema=schema)

    @timeit(text='Uploading df to table took:', logger=LOGGER)
    def to_table_from_df(self, df: pd.DataFrame, table_reference: str, if_exists: str = 'fail', index: bool = False,
                         index_label: Union[str, Sequence, None] = None) -> None:
        """
        Uploads a DataFrame to a table and creates it if it does not exist yet.
        :param df: pandas DataFrame to upload to BigQuery
        :param table_reference: reference of the table to create/append/replace (e.g. "warehouse.trip_fill" or
            "bbc-data-platform.warehouse.trip_fill")
        :param if_exists: action to take if the table already exists. Should be in ('fail', 'append', 'replace')
        :param index: write df index as a column. Uses `index_label` as the column name in the table
        :param index_label: column label for index column(s). If None is given (default) and `index` is True, then the
            index names are used.
        :return: None
        """
        schema, table = self._parse_table_reference(table_reference)
        df.to_sql(table, con=self.conn, schema=schema, if_exists=if_exists, index=index, index_label=index_label)

    @timeit(text='Running query and storing results into table took:', logger=LOGGER)
    def to_table_from_query(self, query: str, table_reference: str, if_exists: str = 'fail') -> None:
        """
        Run a query and load its result into a table.
        :param query: query to be executed
        :param table_reference: reference of the table to create/append/replace (e.g. "table_name" or
            "schema_name.table_name")
        :param if_exists: behaviour to adopt when uploading results to BigQuery. Ignored when `to_table` is not
            provided. If provided, must be in ('fail', 'append', 'replace'). Defaults to 'fail'
        :return: None
        """
        schema, table = self._parse_table_reference(table_reference)
        if not self.exists(table, schema):
            self._execute(f'CREATE TABLE {table_reference} AS {query}')
        else:
            if if_exists == 'fail':
                raise ValueError(f'"{table_reference}" table already exists and if_exists set to {if_exists}')
            elif if_exists == 'replace':
                truncate_statement = 'DELETE FROM' if self.db_type == 'sqlite' else 'TRUNCATE TABLE'
                self._execute(f'{truncate_statement} {table_reference};')
                self._execute(f'INSERT INTO {table_reference} {query};')
            elif if_exists == 'append':
                self._execute(f'INSERT INTO {table_reference} {query};')
            else:
                raise NotImplementedError('"if_exists" argument must be in ("fail", "replace", "append")')


class PostgresWrapper(_DatabaseWrapper):

    def __init__(self, conn):
        super(PostgresWrapper, self).__init__(conn, 'postgres')

    @classmethod
    def from_params(cls, host: str, port: int, db_name: str, user: str, password: str, **_) -> _DatabaseWrapper:
        return _DatabaseWrapper(create_engine(f'postgresql+psycopg2://{user}:{password}@{host}:{port}/{db_name}'),
                                'postgres')


class MysqlWrapper(_DatabaseWrapper):

    def __init__(self, conn):
        super(MysqlWrapper, self).__init__(conn, 'mysql')

    @classmethod
    def from_params(cls, host: str, port: int, db_name: str, user: str, password: str, **_) -> _DatabaseWrapper:
        return _DatabaseWrapper(create_engine(f'mysql+pymysql://{user}:{password}@{host}:{port}/{db_name}'), 'mysql')


class SqliteWrapper(_DatabaseWrapper):

    def __init__(self, conn):
        super(SqliteWrapper, self).__init__(conn, 'sqlite')

    @classmethod
    def from_params(cls, db_filepath: str, **_) -> _DatabaseWrapper:
        return _DatabaseWrapper(create_engine(f'sqlite:///{db_filepath}'), 'sqlite')



import logging
import os
import tempfile
import zipfile
from contextlib import contextmanager
from glob import glob
from typing import Optional
from uuid import uuid4

import yaml

from ._logging import log

LOGGER = logging.getLogger(__name__)


@contextmanager
def ensure_directory_exists(directory_path: str) -> None:
    """
    Context manager and decorator to ensure a directory exists, and create it if it does not.
    :param directory_path: path to the folder
    :return: None
    """
    try:
        LOGGER.debug(f'Creating directory {directory_path}')
        if directory_path not in ['', '.']:
            os.makedirs(directory_path, exist_ok=True)
        yield
    finally:
        pass


@contextmanager
def ignore_exceptions(*exceptions: Exception) -> None:
    """
    Context manager and decorator to ignore Python exceptions. If no exception types are provided, all exceptions will
    be ignored. Should be used cautiously.
    :param exceptions: specific exceptions to be ignored
    :return: None
    """
    try:
        yield
    except tuple(exceptions) or Exception:
        pass


def safe_write(filepath: str, content: str) -> None:
    """
    Write content into a file (existing or not), and create the directory of the file if it does not exist.
    :param filepath: path to the file to be written
    :param content: content to be written in the file
    :return: None
    """
    with ensure_directory_exists(os.path.dirname(filepath)):
        with open(filepath, "w") as f:
            f.write(content)


def read_yaml(path: str) -> dict:
    """
    Read and load a YAML file safely.
    :param path: path to the YAML file to load.
    :return: loaded YAML file as a dict
    """
    return yaml.load(open(path, 'r'), Loader=yaml.SafeLoader)


def get_tmp_dirpath() -> str:
    """
    Generate temporary directory and return its path.
    :return: path to the temporary directory
    """
    dirpath = f'/tmp/{str(uuid4())}'
    with ensure_directory_exists(dirpath):
        return dirpath


def get_tmp_filepath(tmp_dirpath: Optional[str] = None, suffix: str = '') -> str:
    """
    Generate filepath for a file in a temporary folder, which can be specified or not.
    :param tmp_dirpath: optional directory, where the temporary file should be located in, when provided
    :param suffix: optional suffix to add to the filepath
    :return: generated filepath
    """
    filename_suffix = '_' + suffix if suffix else ''
    with tempfile.NamedTemporaryFile(suffix=filename_suffix) as tmp_file:
        tmp_dirpath = tmp_dirpath or os.path.dirname(tmp_file.name)
        suffix = os.path.basename(tmp_file.name)
    return os.path.join(tmp_dirpath, suffix)


def zip_files(input_filepath: str, output_filepath: str, exclude_filepath: str = '', if_exists: Optional[str] = 'fail'):
    """
    Function taking a path (or glob path pattern) and creating a ZIP file with all content matching this path.
    :param input_filepath: input file(s) pathname in the glob unix pathname pattern format. More information here:
        https://docs.python.org/3.7/library/glob.html#module-glob
    :param output_filepath: path to the ZIP file to be created
    :param exclude_filepath: file(s) pathname to be excluded from the ZIP file being created.
    :param if_exists: action to take if the output file already exists. Should be in ('fail', 'replace')
    :return: None
    """
    # If the output file already exists, handle it beforehand
    if os.path.exists(output_filepath):
        if if_exists == 'fail':
            raise FileExistsError(f'File {output_filepath} already exists. Set if_exists to "replace" to overwrite it.')
        elif if_exists == 'replace':
            log(f'File {output_filepath} is being replaced.', logger=LOGGER, level=logging.WARNING)
            os.remove(output_filepath)
        else:
            raise NotImplementedError
    # Create the output file
    input_files = glob(input_filepath, recursive=True)  # Need to be done outside of the with context manager
    excluded_files = glob(exclude_filepath, recursive=True)
    with zipfile.ZipFile(output_filepath, 'w', zipfile.ZIP_DEFLATED) as output:
        for filepath in (set(input_files) - set(excluded_files)):
            log(f'Writing: {filepath}', logger=LOGGER, level=logging.DEBUG)
            output.write(filepath)
    log(f'Successfully zipped "{input_filepath}" into "{output_filepath}"', logger=LOGGER, level=logging.INFO)


def unzip_file(input_filepath: str, output_dirpath: str = '.'):
    """
    Uncompress a ZIP file into a specific directory.
    :param input_filepath: path to the ZIP file to unzip
    :param output_dirpath: path to the directory to unzip the file into. Defaults to current working directory
    :return: None
    """
    file = zipfile.ZipFile(input_filepath, 'r', zipfile.ZIP_DEFLATED)
    file.extractall(path=output_dirpath)
    log(f'Successfully unzipped "{input_filepath}" into "{output_dirpath}"', logger=LOGGER, level=logging.INFO)



import logging
import os
from typing import Optional

from six.moves.urllib.parse import quote

from dslib.utils import ensure_directory_exists

try:
    from google.cloud import storage
    from google.cloud.exceptions import NotFound
except ImportError as e:
    raise ImportError('Please install "google" extra packages.') from e

# Set up logger
logging.getLogger('oauth2client').setLevel(logging.WARNING)  # Silence google authentication
LOGGER = logging.getLogger(__name__)

# Set up special types
BucketType = storage.bucket.Bucket


class StorageWrapper(object):

    def __init__(self, client, default_bucket: Optional[str] = None) -> None:
        if not client.__class__.__name__ == 'Client':
            raise ValueError('"client" argument must be a Google "Client" instance')
        self.client = client
        self.default_bucket_name = default_bucket

    @classmethod
    def from_params(cls, project: str, default_bucket: Optional[str] = None, auth_file: Optional[str] = None):
        if auth_file:
            if not os.path.exists(auth_file):
                raise FileNotFoundError(f'Authentication file does not exist: {auth_file}')
            client = storage.Client.from_service_account_json(auth_file, **{'project': project})
        else:
            client = storage.Client(project=project)
        return StorageWrapper(client, default_bucket)

    def _require_bucket(self, bucket_name: Optional[str] = None) -> BucketType:
        bucket_name = bucket_name or self.default_bucket_name
        if not bucket_name:
            raise ValueError('You need to specify a bucket_name, either at wrapper level or function level')
        return self.client.bucket(bucket_name)

    def upload_file(self, source_filepath: str, destination_filepath: str, bucket_name: Optional[str] = None,
                    replace: bool = False) -> None:
        """
        Upload a file to Google Cloud Storage.
        :param source_filepath: (str) local path of the file to upload
        :param destination_filepath: (str) path to the destination file in the bucket
        :param bucket_name: (str) name of the bucket in which the file should be uploaded (if None, the default bucket
            of the StorageWrapper instance will be used)
        :param replace: (bool) replace the file if it already exists when set to True. Defaults to False
        :return: None
        """
        bucket = self._require_bucket(bucket_name)
        blob = bucket.blob(destination_filepath)
        if replace is False and blob.exists():
            raise FileExistsError(f'gs://{bucket.name}/{destination_filepath} already exists')
        with open(source_filepath, 'rb') as fileobj:
            blob.upload_from_file(fileobj)
        LOGGER.info(f'Uploaded file {source_filepath} to gs://{bucket.name}/{destination_filepath}')

    def download_file(self, source_filepath: str, destination_filepath: str, bucket_name: Optional[str] = None) -> None:
        """
        Downloads a file from Google Cloud Storage.
        :param source_filepath: (str) path of the distant file to download
        :param destination_filepath: (str) local path to store the downloaded file to
        :param bucket_name: (str) bucket name of the file to download (if None, the default bucket of the StorageWrapper
            instance will be used)
        :return: None
        """
        bucket = self._require_bucket(bucket_name)
        blob = bucket.blob(source_filepath)
        with ensure_directory_exists(os.path.dirname(destination_filepath)):
            try:
                blob.download_to_filename(destination_filepath)
            except NotFound as exception:
                raise FileNotFoundError(f'Could not download {source_filepath}') from exception
            except Exception as exception:
                if os.path.exists(destination_filepath):
                    os.remove(destination_filepath)
                LOGGER.error(f'Could not download file gs://{bucket.name}/{source_filepath} to {destination_filepath}')
                raise exception
            else:
                LOGGER.info(f'Downloaded file gs://{bucket.name}/{source_filepath} to {destination_filepath}')

    def delete_file(self, filepath: str, bucket_name: Optional[str] = None) -> None:
        """
        Delete a file from Google Cloud Storage.
        :param filepath: path of the distant file to delete
        :param bucket_name: bucket name of the file to delete (if None, the default bucket of the StorageWrapper
            instance will be used)
        :return: None
        """
        bucket = self._require_bucket(bucket_name)
        blob = bucket.blob(filepath)
        try:
            blob.delete()
            LOGGER.info(f'Deleted file gs://{bucket.name}/{filepath}')
        except NotFound:
            LOGGER.warning(f'Could not delete gs://{bucket.name}/{filepath}: file file not found')

    def exists(self, filepath: str, bucket_name: Optional[str] = None) -> bool:
        """
        Check the existence of a file from Google Cloud Storage.
        :param filepath: path of the distant file to check the existence of
        :param bucket_name: bucket name of the file to check the existence of (if None, the default bucket of the
            StorageWrapper instance will be used)
        :return: True if the file exists, False otherwise
        """
        bucket = self._require_bucket(bucket_name)
        blob = bucket.blob(filepath)
        exists = blob.exists()
        LOGGER.debug(f"File {filepath} {'exists' if exists else 'does not exist'}")
        return exists

    def get_public_url(self, filepath: str, bucket_name: Optional[str] = None) -> str:
        """
        Make a distant file from Google Cloud Storage public and get its public URL.
        :param filepath: path of the distant file to make public and to get the URL from
        :param bucket_name: bucket name of the file to make public and get the URL of (if None, the default bucket of
            the StorageWrapper instance will be used)
        :return: public URL of the file
        """
        bucket = self._require_bucket(bucket_name)
        blob = bucket.blob(filepath)
        blob.make_public()
        LOGGER.warning(f'gs://{bucket.name}/{filepath} was made public')
        return f'{"https://storage.googleapis.com"}/{bucket.name}/{quote(filepath)}'



from datetime import date, timedelta, datetime

from dateutil import tz


def midpoint_date(first_date: date, second_date: date, mid_at: float = 0.5) -> date:
    """
    Returns the midpoint date between two dates
    :param first_date: first date in chronological order
    :param second_date: second date in chronological order
    :param mid_at: change the position of the mid date (e.g. put it at two third rather than at the exact midpoint)
    :return: midpoint date
    """
    return first_date + (second_date - first_date) * mid_at


def date_range(start_date: date, end_date: date):
    """
    Returns a generator of all ordered dates between `start_date` and `end_date`.
    :param start_date: first date of the `date_range` generator
    :param end_date: last date of the `date_range` generator
    :return: generator of dates
    """
    return (start_date + timedelta(days=i) for i in range((end_date - start_date).days + 1))


def change_timezone(timestamp: datetime, timezone_name: str):
    """
    Takes a timestamp and changes its timezone to a provided timezone_name. If the provided timestamp is not timezone-
    aware, it is considered UTC.
    :param timestamp: timezone-aware or not (in which case it is considered UTC)
    :param timezone_name: name of the timezone to change to (e.g. "America/Sao_Paulo")
    :return: timezone-aware timestamp
    """
    timezone = tz.gettz(timezone_name)
    if timezone is not None:
        return timestamp.astimezone(timezone)
    else:
        raise ValueError(f'Could not find timezone with name: {timezone_name}')


def force_timezone_onto_timestamp(timestamp: datetime, timezone_name: str):
    """
    Takes a timestamp (timezone-aware or not) and forces a specific timezone to it. For instance, applying this function
    to a "1pm in UTC" timestamp with a "Europe/Paris" timezone_name would result in a "1pm, CET" timestamp.
    :param timestamp: timestamp (timezone-aware or not)
    :param timezone_name: name of the timezone to force onto the timestamp (e.g. "America/Sao_Paulo")
    :return: timezone-aware timestamp
    """
    timezone = tz.gettz(timezone_name)
    if timezone is not None:
        return timestamp.replace(tzinfo=timezone)
    else:
        raise ValueError(f'Could not find timezone with name: {timezone_name}')



import logging
import os
import sys
from contextlib import contextmanager
from time import time
from typing import Optional

from humanfriendly import format_timespan


def log(message: str, logger: Optional[logging.Logger] = None, level: int = logging.INFO) -> None:
    """
    Logs a message at the provided level, or prints it if no logger is provided.
    :param message: message to be logged
    :param logger: logger to be used for logging. If None, then "print" function is used
    :param level: logger level to be used when a logger is provided. Defaults to logging.INFO
    :return: None
    """
    if logger:
        logger.log(msg=message, level=level)
    else:
        print(message)


def configure_logging(level: int = logging.INFO, detailed: bool = False,
                      logger: Optional[logging.Logger] = None) -> None:
    """
    Configure basic logging.
    :param level: default level to be set for the logging configuration. Defaults to logging.INFO
    :param detailed: whether the log format should be detailed of not. Defaults to False
    :param logger: whether the log format should be detailed of not. Defaults to False
    :return: None
    """
    # Get root logger
    root_logger = logging.getLogger()
    root_logger.setLevel(level)
    # Create formatter
    if detailed:
        date_format = '%Y-%m-%d %H:%M:%S'
        log_format = '[%(asctime)s][%(levelname)-8s][%(name)-12s]: %(message)s'
    else:
        date_format = '%H:%M:%S'
        log_format = '%(asctime)s %(levelname)-8s %(message)s'
    formatter = logging.Formatter(log_format, datefmt=date_format)
    # Create sh handler
    handler_sh = logging.StreamHandler(sys.stdout)
    handler_sh.setLevel(level)
    # Add formatter to handler, and handler to root logger
    handler_sh.setFormatter(formatter)
    if root_logger.hasHandlers():
        root_logger.handlers = []
    root_logger.addHandler(handler_sh)
    # Confirm logging was set
    log('Logging was successfully set', logger=logger)


@contextmanager
def timeit(logger: Optional[logging.Logger] = None, text: str = 'Processing took'):
    """
    Context manager and decorator to print or log processing time.
    :param logger: if provided, will be used to log processing time. Otherwise, "print" function is used
    :param text: text to log (will be followed by the time to process, e.g. " 3.2 seconds")
    :return: None
    """
    start = time()
    try:
        yield
    except Exception as e:
        raise e
    else:
        time_span = time() - start
        detailed = time_span < 0.1
        time_span_formatted = format_timespan(time_span, detailed=detailed) or '0 milliseconds'
        log(f'{text.strip()} {time_span_formatted}', logger)


@contextmanager
def suppress_stdout():
    """
    Context manager and decorator to hide stdout of a part of your code. Can be used on prints.
    :return: None
    """
    original_stdout = sys.stdout
    try:
        sys.stdout = open(os.devnull, 'w')
        yield
    finally:
        sys.stdout.close()
        sys.stdout = original_stdout


class deep_suppress_stdout_stderr(object):
    """
    Context manager doing a "deep suppression" of stdout and stderr, i.e. will suppress all print, even if the print
    originates in a compiled C/Fortran sub-function. This will not suppress raised exceptions, since exceptions are
    printed to stderr just before a script exits, and after the context manager has exited.
    Copied from: https://stackoverflow.com/questions/11130156/suppress-stdout-stderr-print-from-python-functions
    """
    def __init__(self):
        # Open a pair of null files
        self.null_fds = [os.open(os.devnull, os.O_RDWR) for x in range(2)]
        # Save the actual stdout (1) and stderr (2) file descriptors.
        self.save_fds = [os.dup(1), os.dup(2)]

    def __enter__(self):
        # Assign the null pointers to stdout and stderr.
        os.dup2(self.null_fds[0], 1)
        os.dup2(self.null_fds[1], 2)

    def __exit__(self, *_):
        # Re-assign the real stdout/stderr back to (1) and (2)
        os.dup2(self.save_fds[0], 1)
        os.dup2(self.save_fds[1], 2)
        # Close the null files
        for fd in self.null_fds + self.save_fds:
            os.close(fd)



import logging
import os
from mimetypes import guess_type
from typing import Optional, List

try:
    from oauth2client.client import GoogleCredentials
    from oauth2client.service_account import ServiceAccountCredentials
    from googleapiclient.discovery import build, Resource
    from googleapiclient.http import MediaFileUpload
except ImportError as e:
    raise ImportError('Please install "google" extra packages.') from e

LOGGER = logging.getLogger(__name__)


class DriveError(Exception):
    pass


class DriveWrapper(object):

    scopes = [
        'https://www.googleapis.com/auth/drive',
    ]

    def __init__(self, client: Resource, default_folder_id: Optional[str] = None):
        if not client.__class__.__name__ == 'Resource':
            raise ValueError('"client" argument must be a Google Drive "Resource" instance')
        self.client = client
        self.default_folder_id = default_folder_id

    @classmethod
    def from_params(cls, default_folder_id: Optional[str] = None, auth_file: str = None):
        if auth_file:
            if not os.path.exists(auth_file):
                raise FileNotFoundError(f'Authentication file does not exist: {auth_file}')
            credentials = ServiceAccountCredentials.from_json_keyfile_name(auth_file, scopes=cls.scopes)
        else:
            credentials = GoogleCredentials.get_application_default().create_scoped(cls.scopes)
        client = build('drive', version='v3', credentials=credentials, cache_discovery=False)
        return cls(client, default_folder_id=default_folder_id)

    def _require_folder_id(self, folder_id: Optional[str] = None) -> str:
        folder_id = folder_id or self.default_folder_id
        if not folder_id:
            raise ValueError('You need to specify a "folder_id", either at wrapper level or function level')
        return folder_id

    def _find_all(self, filename: str, folder_id: str) -> List[dict]:
        query = f"'{folder_id}' in parents and name = '{filename}'"
        results = self.client.files().list(q=query).execute()
        return results.get('files', [])

    def file_exists(self, filename: str, folder_id: Optional[str] = None) -> bool:
        """
        Returns True if a file named like filename is already present in the folder.
        :param filename: name of the file to check the existence of
        :param folder_id: unique ID of the distant folder. If not specified, the default folder ID is used
        """
        folder_id = self._require_folder_id(folder_id)
        files = self._find_all(filename, folder_id)
        return len(files) > 0

    def find_file(self, filename: str, folder_id: Optional[str] = None) -> Optional[str]:
        """
        Find the ID of a file by its name. Raises if multiple files have the searched name.
        :param filename: (str) name of the files to search for (several files can have the same name).
        :param folder_id: (str) unique ID of the distant folder. If not specified, the default folder ID is used
        :return: id of the file when it is found, None otherwise
        """
        folder_id = self._require_folder_id(folder_id)
        file_ids = [item['id'] for item in self._find_all(filename, folder_id)]
        if len(file_ids) == 0:
            return None
        elif len(file_ids) > 1:
            raise DriveError(f'Multiple files with name "{filename}" in folder "{folder_id}": {", ".join(file_ids)}')
        else:
            return file_ids[0]

    def delete_file(self, filename: str, folder_id: Optional[str] = None) -> None:
        """
        Deletes a file with a specific name, in the specified folder or the default one. If several files have the same
        name, they will all be deleted.
        :param filename: (str) name of the file to delete
        :param folder_id: (str) unique ID of the distant folder. If not specified, the default folder ID is used
        :return: None
        """
        folder_id = self._require_folder_id(folder_id)
        files = self._find_all(filename, folder_id)
        if not files:
            LOGGER.info(f'Could not find "{filename}" in directory "{folder_id}"')
        else:
            LOGGER.info(f'Deleting {len(files)} file{"s" if len(files) > 1 else ""}...')
        for file in files:
            self.client.files().delete(fileId=file['id']).execute()

    def give_ownership(self, file_id: str, new_owner_email: str) -> None:
        """
        Gives a user ownership rights on a specific file.
        :param file_id: unique ID of the file. Can be found using "find_file" method
        :param new_owner_email: email address of the user to give ownership rights to
        :return: None
        """
        batch = self.client.new_batch_http_request()
        user_permission = {
            'type': 'user',
            'role': 'owner',
            'emailAddress': new_owner_email
        }
        client_permission = self.client.permissions().create(
            fileId=file_id,
            body=user_permission,
            fields='id'
        )
        batch.add(client_permission)
        batch.execute()
        LOGGER.info(f"Ownership of file {file_id} was given to {new_owner_email}")

    def _upload(self, filename: str, folder_id: Optional[str] = None, mimetype: Optional[str] = None) -> str:
        folder_id = self._require_folder_id(folder_id)
        file_metadata = {
            'name': filename,
            'mimeType': mimetype,
            'field': 'id',
            'parents': [folder_id]
        }
        mimetype = mimetype or guess_type(filename)[0]
        media = MediaFileUpload(filename, mimetype=mimetype, resumable=True)
        created_file = self.client.files().create(body=file_metadata, media_body=media, fields='id').execute()
        file_id = created_file.get('id')
        LOGGER.info(f'File "{filename}" with ID "{file_id}" successfully uploaded in folder "{folder_id}"')
        return file_id

    def upload_file(self, filename: str, folder_id: Optional[str] = None, mimetype: Optional[str] = None,
                    if_exists: str = 'fail') -> str:
        """
        Send a file from current directory to Google Drive.
        TODO: Support uploading files from other directories than working directory
        :param filename: (str) local filename to send (must be in working directory)
        :param folder_id: (str) target folder ID. If not specified, the default folder ID is used
        :param mimetype: (str) MIME type of the file to send. If None, then auto-detect is used
        :param if_exists: (str) action to take if this filename already exists. Should be in ('fail', 'replace')
        :return: file_id of the uploaded file
        """
        folder_id = self._require_folder_id(folder_id)
        if self.file_exists(filename, folder_id):
            if if_exists == 'fail':
                raise DriveError('File already exists and "if_exists" set to "fail". Upload cancelled')
            elif if_exists == 'replace':
                LOGGER.warning(f'File "{filename}" will be replaced')
                self.delete_file(filename, folder_id)
            else:
                raise NotImplementedError('"if_exists" argument must be in ("fail", "duplicate")')
        LOGGER.info(f'Adding new file "{filename}"...')
        file_id = self._upload(filename, folder_id, mimetype)
        return file_id



from git import Repo
from git.exc import GitCommandError

from version import __version__


repo = Repo()
try:
    print(f'Create git tag "{__version__}"')
    new_tag = repo.create_tag(__version__, message=f'Automatic tag "{__version__}"')
except GitCommandError as exception:
    raise ValueError(f'Invalid version "{__version__}": change it in dslib/__init__.py') from exception
print('Pushing local changes to remote git repository')
repo.remotes.origin.push(new_tag)



import logging
import os
import string
from itertools import islice
from typing import Optional

try:
    import gspread
    import pandas as pd
    from gspread_dataframe import set_with_dataframe
    from oauth2client.service_account import ServiceAccountCredentials
    from oauth2client.client import GoogleCredentials
except ImportError as e:
    raise ImportError('Please install "google" extra packages.') from e

LOGGER = logging.getLogger(__name__)


class SheetError(Exception):
    pass


def colname_generator():
    # First 26 colnames are ASCII uppercase characters
    for letter in string.ascii_uppercase:
        yield letter
    # Then it's AA, AB, AC, ..., BA, BB, BC, ...
    for letter_1 in string.ascii_uppercase:
        for letter_2 in string.ascii_uppercase:
            yield letter_1 + letter_2


class SpreadsheetWrapper(object):

    scopes = [
        'https://spreadsheets.google.com/feeds',
        'https://www.googleapis.com/auth/drive'
    ]

    def __init__(self, client: gspread.client.Client, default_spreadsheet_id: Optional[str] = None):
        if not client.__class__.__name__ == 'Client':
            raise ValueError('"client" argument must be a Google Spreadsheet "Client" instance')
        self.client = client
        self.default_spreadsheet_id = default_spreadsheet_id

    @classmethod
    def from_params(cls, default_spreadsheet_id: Optional[str] = None, auth_file: str = None):
        if auth_file:
            if not os.path.exists(auth_file):
                raise FileNotFoundError(f'Authentication file does not exist: {auth_file}')
            credentials = ServiceAccountCredentials.from_json_keyfile_name(auth_file, scopes=cls.scopes)
        else:
            credentials = GoogleCredentials.get_application_default().create_scoped(cls.scopes)
        client = gspread.authorize(credentials)
        return cls(client, default_spreadsheet_id=default_spreadsheet_id)

    def _require_spreadsheet(self, spreadsheet_id: Optional[str] = None) -> gspread.Spreadsheet:
        spreadsheet_id = spreadsheet_id or self.default_spreadsheet_id
        if not spreadsheet_id:
            raise ValueError('You need to specify a bucket_name, either at wrapper level or function level')
        return self.client.open_by_key(spreadsheet_id)

    @staticmethod
    def _get_worksheet(spreadsheet: gspread.Spreadsheet, worksheet_title: str) -> gspread.Worksheet:
        if not worksheet_title:
            return spreadsheet.get_worksheet(0)
        else:
            return spreadsheet.worksheet(worksheet_title)

    @staticmethod
    def _get_nth_colname(n):
        return next(islice(colname_generator(), n, n+1))

    def worksheet_exists(self, worksheet_title: str, spreadsheet_id: Optional[str] = None) -> bool:
        """
        Check whether a specific worksheet already exists or not.
        :param worksheet_title: (str) title of the worksheet to check the existence of
        :param spreadsheet_id: (str) name of the spreadsheet to look into (if None, the default spreadsheet of the
            SpreadsheetWrapper instance will be used)
        :return: bool
        """
        sh = self._require_spreadsheet(spreadsheet_id)
        # Get list of worksheet titles and check if it is in it
        worksheet_titles = [worksheet.title for worksheet in sh.worksheets()]
        return worksheet_title in worksheet_titles

    def create_worksheet(self, worksheet_title: str, spreadsheet_id: Optional[str] = None, nb_rows: int = 1000,
                         nb_columns: int = 26, index: Optional[int] = None) -> None:
        """
        Create a new worksheet, identified by its title.
        :param worksheet_title: (str) title of the worksheet to create
        :param spreadsheet_id: (str) name of the spreadsheet to create the worksheet into (if None, the default
            spreadsheet of the SpreadsheetWrapper instance will be used)
        :param nb_rows: (int) number of rows of the empty spreadsheet
        :param nb_columns: (int) number of columns of the empty spreadsheet
        :param index: (int) position of the sheet. If None, the new spreadsheet will be positioned last
        :return: None
        """
        LOGGER.info(f'Creating empty worksheet {worksheet_title}...')
        sh = self._require_spreadsheet(spreadsheet_id)
        sh.add_worksheet(worksheet_title, nb_rows, nb_columns, index)

    def delete_worksheet(self, worksheet_title: str, spreadsheet_id: Optional[str] = None) -> None:
        """
        Delete a worksheet, identified by its title.
        :param worksheet_title: (str) title of the worksheet to delete
        :param spreadsheet_id: (str) name of the spreadsheet the worksheet to delete is in (if None, the default
            spreadsheet of the StorageWrapper instance will be used)
        :return: None
        """
        LOGGER.info(f'Deleting worksheet {worksheet_title}...')
        sh = self._require_spreadsheet(spreadsheet_id)
        worksheet = self._get_worksheet(sh, worksheet_title)
        sh.del_worksheet(worksheet)

    def clear_worksheet(self, worksheet_title: str, spreadsheet_id: Optional[str] = None) -> None:
        """
        Clear all values from a worksheet, identified by its title.
        :param worksheet_title: (str) title of the worksheet to clear
        :param spreadsheet_id: (str) name of the spreadsheet the worksheet to clear is in (if None, the default
            spreadsheet of the StorageWrapper instance will be used)
        :return: None
        """
        LOGGER.info(f'Clearing spreadsheet {worksheet_title}...')
        # Load the worksheet to empty
        sh = self._require_spreadsheet(spreadsheet_id)
        worksheet = self._get_worksheet(sh, worksheet_title)
        # Add an empty column, remove all N columns but this one, and recreate N-1 empty columns
        sh.values_clear("'{worksheet_title}'!A1:{last_colname}{last_rownumber}".format(
            worksheet_title=worksheet_title,
            last_colname=self._get_nth_colname(worksheet.col_count),
            last_rownumber=worksheet.row_count
        ))

    def worksheet_to_df(self, worksheet_title: str, spreadsheet_id: Optional[str] = None) -> pd.DataFrame:
        """
        Read the data from a specific worksheet and return it as a pandas DataFrame.
        :param worksheet_title: (str) title of the worksheet to read
        :param spreadsheet_id: (str) name of the spreadsheet the worksheet to read is in (if None, the default
            spreadsheet of the StorageWrapper instance will be used)
        :return: pandas DataFrame with the worksheet's data
        """
        sh = self._require_spreadsheet(spreadsheet_id)
        # Download data and format it
        data = self._get_worksheet(sh, worksheet_title).get_all_records()
        return pd.DataFrame.from_records(data)

    def df_to_worksheet(self, df: pd.DataFrame, worksheet_title: str, spreadsheet_id: Optional[str] = None,
                        if_exists: str = 'fail', include_index=False, include_header=True) -> None:
        """
        Upload a pandas DataFrame to a specific worksheet, identified by its title.
        :param df: pandas DataFrame to be uploaded to the worksheet
        :param worksheet_title: (str) title of the worksheet to upload the data to
        :param spreadsheet_id: (str) name of the spreadsheet the worksheet to upload the data to is in (if None, the
            default spreadsheet of the StorageWrapper instance will be used)
        :param if_exists: action to take if the worksheet already exists. Should be in ('fail', 'replace')
        :param include_index: (bool) if True, include the DataFrame's index as an additional column. Defaults to False.
        :param include_header: (bool) if True, add a header row or rows before data with column names. Defaults to True.
        :return: None
        """
        # If the worksheet does not exist, create it
        if not self.worksheet_exists(worksheet_title, spreadsheet_id):
            self.create_worksheet(worksheet_title, spreadsheet_id)
        # If it exists, adopt the behavior specified in the "if_exists" argument
        else:
            if if_exists == 'fail':
                raise SheetError(f'Spreadsheet already has a worksheet "{worksheet_title}".'
                                 f'Set "if_exists" to "replace" to clear it before loading data.')
            elif if_exists == 'replace':
                self.clear_worksheet(worksheet_title, spreadsheet_id)
            else:
                raise NotImplementedError('"if_exists" argument must be in ("fail", "replace")')
        # Pre-process data (fill null values)
        sh = self._require_spreadsheet(spreadsheet_id)
        worksheet = self._get_worksheet(sh, worksheet_title)
        set_with_dataframe(worksheet, df, include_index=include_index, include_column_header=include_header)



import os

import setuptools

from version import __version__

CURRENT_FILEPATH = os.path.abspath(os.path.dirname(__file__))
VERSION_FILENAME = 'version.py'


setuptools.setup(
    name='bbc-dslib',
    version=__version__,
    author='Raphaël Berly',
    author_email='raphael.berly@blablacar.com',
    description='A lib for the Data Science team at Blablacar',
    license='closed',
    url="https://github.com/blablacar/data-dslib",
    packages=setuptools.find_packages(),
    python_requires='>=3.6',
    install_requires=['humanfriendly', 'pyyaml', 'jinja2'],
    extras_require={
        'database': ['sqlalchemy', 'psycopg2-binary', 'PyMySQL', 'pandas==1.1.3'],
        'facebook': ['fbprophet'],
        'google': [
            'google-cloud-bigquery==1.28.0',
            'google-cloud-bigquery-storage==0.8.0',
            'google-cloud-storage==1.31.2',
            'google-api-python-client==1.12.3',
            'pyarrow==1.0.1',
            'oauth2client==4.1.3',
            'grpcio==1.32.0',
            'pandas==1.1.3',
            'pandas-gbq==0.14.0',
            'tqdm==4.50.2',
            'gspread==3.6.0',
            'gspread-dataframe==3.0.6'
        ],
        'science': [
            'scikit-learn==0.23.2',
            'numpy==1.19.3',
            'matplotlib==3.2.1',
            'dill==0.3.1.1',
            'pandas==1.1.3'
        ],
        'testing': ['pytest', 'pytest-cov', 'coverage', 'mock', 'testfixtures'],
    }
)



import os
from typing import Optional

from jinja2 import Environment, FileSystemLoader, StrictUndefined


def render_file(filepath: str, params: Optional[dict] = None, **kwargs) -> str:
    """
    Render a jinja template file. Includes and imports root is set to the directory component of ``filepath``.
    :param filepath: The path to the template file
    :param params: The parameters to bind to the template
    :param kwargs: The other parameters to bind to the template
    :return: The rendered template
    :raise: jinja2.UndefinedError if a parameter is undefined
    """
    folder = os.path.dirname(filepath)
    filepath = os.path.basename(filepath)
    params = params or {}
    environment = Environment(loader=FileSystemLoader(folder), undefined=StrictUndefined)
    return environment.get_template(filepath).render(params, **kwargs)


def render_string(string: str, params: Optional[dict] = None, **kwargs) -> str:
    """
    Render a jinja template string.
    :param string: The template string
    :param params: The parameters to bind to the template
    :param kwargs: The other parameters to bind to the template
    :return: The rendered template
    :raise: jinja2.UndefinedError if a parameter is undefined
    """
    params = params or {}
    environment = Environment(undefined=StrictUndefined)
    return environment.from_string(string).render(params, **kwargs)
